using System;
using System.Collections.Generic;
using System.Linq;

namespace Encompass
{
    internal class ComponentManager
    {
        private readonly DrawLayerManager drawLayerManager;

        private readonly Dictionary<Guid, Type> componentIDToType = new Dictionary<Guid, Type>();
        private readonly Dictionary<Guid, IComponent> IDToComponent = new Dictionary<Guid, IComponent>();
        private readonly Dictionary<Guid, List<Guid>> entityIDToComponentIDs = new Dictionary<Guid, List<Guid>>();
        private readonly Dictionary<Guid, Guid> componentIDToEntityID = new Dictionary<Guid, Guid>();

        private readonly Dictionary<Type, List<Guid>> typeToComponentIDs = new Dictionary<Type, List<Guid>>();

        private readonly List<Guid> activeComponents = new List<Guid>();
        private readonly List<Guid> inactiveComponents = new List<Guid>();

        //shared references with EntityManager
        private readonly HashSet<Guid> entitiesWithAddedComponents;
        private readonly HashSet<Guid> entitiesWithRemovedComponents;

        public ComponentManager(
            DrawLayerManager drawLayerManager,
            HashSet<Guid> entitiesWithAddedComponents,
            HashSet<Guid> entitiesWithRemovedComponents
        )
        {
            this.drawLayerManager = drawLayerManager;
            this.entitiesWithAddedComponents = entitiesWithAddedComponents;
            this.entitiesWithRemovedComponents = entitiesWithRemovedComponents;
        }

        internal void RegisterEntity(Guid entityID)
        {
            entityIDToComponentIDs.Add(entityID, new List<Guid>());
        }

        internal Guid AddComponent<TComponent>(Guid entityID, TComponent component) where TComponent : struct, IComponent
        {
            var componentID = Guid.NewGuid();

            IDToComponent[componentID] = component;
            componentIDToType[componentID] = typeof(TComponent);

            if (!typeToComponentIDs.ContainsKey(typeof(TComponent)))
            {
                typeToComponentIDs.Add(typeof(TComponent), new List<Guid>());
            }

            typeToComponentIDs[typeof(TComponent)].Add(componentID);

            entityIDToComponentIDs[entityID].Add(componentID);
            componentIDToEntityID[componentID] = entityID;

            inactiveComponents.Add(componentID);
            Activate(componentID);

            entitiesWithAddedComponents.Add(entityID);

            return componentID;
        }

        internal Guid AddDrawComponent<TComponent>(Guid entityID, TComponent component, int layer = 0) where TComponent : struct, IComponent
        {
            var componentID = AddComponent(entityID, component);
            drawLayerManager.RegisterComponentWithLayer(componentID, layer);
            return componentID;
        }

        internal IEnumerable<Guid> GetComponentIDsByEntityID(Guid entityID)
        {
            return entityIDToComponentIDs.ContainsKey(entityID) ?
                entityIDToComponentIDs[entityID] :
                Enumerable.Empty<Guid>();
        }

        internal IEnumerable<ValueTuple<Guid, IComponent>> GetComponentsByEntity(Guid entityID)
        {
            return GetComponentIDsByEntityID(entityID).Intersect(activeComponents).Select((id) => new ValueTuple<Guid, IComponent>(id, IDToComponent[id]));
        }

        internal IEnumerable<ValueTuple<Guid, TComponent>> GetActiveComponentsByType<TComponent>() where TComponent : struct, IComponent
        {
            return typeToComponentIDs.ContainsKey(typeof(TComponent)) ?
                typeToComponentIDs[typeof(TComponent)].Intersect(activeComponents).Select((id) => new ValueTuple<Guid, TComponent>(id, (TComponent)IDToComponent[id])) :
                Enumerable.Empty<ValueTuple<Guid, TComponent>>();
        }

        internal IEnumerable<ValueTuple<Guid, IComponent>> GetActiveComponentsByType(Type type)
        {
            return typeToComponentIDs.ContainsKey(type) ?
                typeToComponentIDs[type].Intersect(activeComponents).Select((id) => new ValueTuple<Guid, IComponent>(id, IDToComponent[id])) :
                Enumerable.Empty<ValueTuple<Guid, IComponent>>();
        }

        internal ValueTuple<Guid, TComponent> GetActiveComponentByType<TComponent>() where TComponent : struct, IComponent
        {
            return GetActiveComponentsByType<TComponent>().Single();
        }

        internal IEnumerable<ValueTuple<Guid, TComponent>> GetComponentsByEntityAndType<TComponent>(Guid entityID) where TComponent : struct, IComponent
        {
            var entityComponentsByType = GetComponentsByEntity(entityID).Where((pair) => componentIDToType[pair.Item1] == typeof(TComponent)).Select((pair) => new ValueTuple<Guid, TComponent>(pair.Item1, (TComponent)pair.Item2));
            var activeComponentsByType = GetActiveComponentsByType<TComponent>();
            return activeComponentsByType.Intersect(entityComponentsByType);
        }

        internal IEnumerable<ValueTuple<Guid, IComponent>> GetComponentsByEntityAndType(Guid entityID, Type type)
        {
            var entityComponents = GetComponentsByEntity(entityID);
            var activeComponentsByType = GetActiveComponentsByType(type);
            return entityComponents.Intersect(activeComponentsByType);
        }

        internal IEnumerable<Type> GetAllComponentTypesOfEntity(Guid entityID)
        {
            return GetComponentIDsByEntityID(entityID).Select((id) => componentIDToType[id]);
        }

        internal bool EntityHasComponentOfType<TComponent>(Guid entityID) where TComponent : struct, IComponent
        {
            return GetComponentsByEntityAndType<TComponent>(entityID).Any();
        }

        internal bool EntityHasComponentOfType(Guid entityID, Type type)
        {
            return GetComponentsByEntityAndType(entityID, type).Any();
        }

        internal IComponent GetComponentByID(Guid componentID)
        {
            return IDToComponent[componentID];
        }

        internal Type GetComponentTypeByID(Guid componentID)
        {
            return componentIDToType[componentID];
        }

        internal Guid GetEntityIDByComponentID(Guid componentID)
        {
            return componentIDToEntityID[componentID];
        }

        internal void UpdateComponent<TComponent>(Guid componentID, TComponent newComponentValue) where TComponent : struct, IComponent
        {
            IDToComponent[componentID] = newComponentValue;
        }

        internal void RemoveAllComponentsFromEntity(Guid entityID)
        {
            var componentIDs = entityIDToComponentIDs[entityID];

            for (int i = componentIDs.Count - 1; i >= 0; i--)
            {
                Remove(componentIDs[i]);
            }
        }

        internal void Activate(Guid componentID)
        {
            if (inactiveComponents.Remove(componentID))
            {
                activeComponents.Add(componentID);
            }

            var entityID = GetEntityIDByComponentID(componentID);
            entitiesWithAddedComponents.Add(entityID);
        }

        internal void Deactivate(Guid componentID)
        {
            if (activeComponents.Remove(componentID))
            {
                inactiveComponents.Add(componentID);
            }

            var entityID = GetEntityIDByComponentID(componentID);
            entitiesWithRemovedComponents.Add(entityID);
        }

        internal void Remove(Guid componentID)
        {
            var component = IDToComponent[componentID];
            var type = componentIDToType[componentID];

            activeComponents.Remove(componentID);
            inactiveComponents.Remove(componentID);

            var entityID = componentIDToEntityID[componentID];
            if (entityIDToComponentIDs.ContainsKey(entityID))
            {
                entityIDToComponentIDs[entityID].Remove(componentID);
            }

            IDToComponent.Remove(componentID);
            componentIDToType.Remove(componentID);
            componentIDToEntityID.Remove(componentID);
            typeToComponentIDs[type].Remove(componentID);

            drawLayerManager.UnRegisterComponentWithLayer(componentID);

            entitiesWithRemovedComponents.Add(entityID);
        }

        public void RegisterDestroyedEntity(Guid entityID)
        {
            entityIDToComponentIDs.Remove(entityID);
        }
    }
}
